import random
import numpy as np
from shapley.solution_concept import SolutionConcept

class PermutationSampler(SolutionConcept):
    r"""Permutation sampler to solve a block of weighted voting games. The solver
    samples random permutations of the players uniformly. In each permutation we
    identify the pivotal voter. We use the empirical probability of becoming the
    pivotal player as the Shapley value estimate. For details see this paper: 
    `"Bounding the Estimation Error of Sampling-based Shapley Value Approximation." 
    <https://arxiv.org/abs/1306.4265>`_

    Args:
        permutations (int): Number of permutations. The default is 1000.
    """    
    def __init__(self, permutations: int=1000):
        self.permutations = permutations

    def _setup(self, W: np.ndarray):
        """Creating an empty Shapley value matrix and a player pool."""
        self._Phi = np.zeros(W.shape)
        self._indices = [i for i in range(W.shape[1])]

    def _run_permutations(self, W: np.ndarray, q: float):
        """Creating Monte Carlo permutations and finding the marginal voter."""
        for _ in range(self.permutations):
            random.shuffle(self._indices)
            W_perm = W[:, self._indices]
            cum_sum = np.cumsum(W_perm, axis=1)
            pivotal = np.array(self._indices)[np.argmax(cum_sum>q, axis=1)]
            self._Phi[np.arange(W.shape[0]), pivotal] += 1.0
        self._Phi = self._Phi/self.permutations


    def solve_game(self, W: np.ndarray, q: float):
        r"""Solving the weigted voting game(s).

        Args:
            W (Numpy array): An :math:`n \times m` matrix of voting weights for the :math:`n` games with :math:`m` players.
            q (float): Quota in the games.
        """
        self._check_quota(q)
        self._setup(W)
        self._run_permutations(W, q)
        self._run_sanity_check(W, self._Phi)
        self._set_average_shapley()
        self._set_shapley_entropy()

    def get_solution(self) -> np.ndarray:
        r"""Returning the solution.

        Return Types:
            Phi (Numpy array): Approximate Shapley matrix of players in the game(s) with size :math:`n \times m`.
        """
        return self._Phi



import itertools
import numpy as np
from shapley.solution_concept import SolutionConcept

class ExactEnumeration(SolutionConcept):
    r"""Exact enumeration of all permutations and finding the pivotal voters. It
    is designed with a generator of the permutations. For details see this paper: 
    `"A Value for N-Person Games." <https://www.rand.org/pubs/papers/P0295.html>`_
    """

    def _setup(self, W: np.ndarray):
        """Creating an empty Shapley value matrix and a player pool."""
        self._Phi = np.zeros(W.shape)
        self._indices = [i for i in range(W.shape[1])]
        self.permutations = 0

    def _run_permutations(self, W: np.ndarray, q: float):
        """Creating Monte Carlo permutations and finding the marginal voter."""
        for perm in itertools.permutations(self._indices):
            self.permutations = self.permutations + 1
            indices = list(perm)
            W_perm = W[:, indices]
            cum_sum = np.cumsum(W_perm, axis=1)
            pivotal = np.array(indices)[np.argmax(cum_sum>q, axis=1)]
            self._Phi[np.arange(W.shape[0]), pivotal] += 1.0
        self._Phi = self._Phi/self.permutations


    def solve_game(self, W: np.ndarray, q: float):
        r"""Solving the weigted voting game(s).

        Args:
            W (Numpy array): An :math:`n \times m` matrix of voting weights for the :math:`n` games with :math:`m` players.
            q (float): Quota in the games.
        """
        self._check_quota(q)
        self._setup(W)
        self._run_permutations(W, q)
        self._run_sanity_check(W, self._Phi)
        self._set_average_shapley()
        self._set_shapley_entropy()

    def get_solution(self) -> np.ndarray:
        r"""Returning the solution.

        Return Types:
            Phi (Numpy array): Approximate Shapley matrix of players in the game(s) with size :math:`n \times m`.
        """
        return self._Phi



import random
import numpy as np
from numba import jit
from scipy.stats import norm
from shapley.solution_concept import SolutionConcept


@jit
def create_integrand_vectors(player_count, q, w):
    """
    Creating the vector of standard deviations and integrand limits.
    :param player_count (int): Number of players in the game.
    :param q (float): Quota to win the game.
    :param w (float): Weight of the player in the game.
    
    :return a_s: Vector of lower integrand limits.
    :return b_s: Vector of upper integrand limits.
    """
    a_s = (q-w)/np.linspace(1, player_count-1, player_count-1)
    b_s = (q-10**-20)/np.linspace(1, player_count-1, player_count-1)
    return a_s, b_s


@jit
def create_standard_deviation_vector(var, player_count):
    """
    Creating the vector of standard deviations and integrand limits.
    :param var (float): Variance of the weights in the game.
    :param player_count (int): Number of players in the game.
    
    :return sigma_s: Vector of standard deviations for the game.
    """
    sigma_s = np.power(var/np.linspace(1, player_count-1, player_count-1), 0.5)
    return sigma_s

class ExpectedMarginalContributions(SolutionConcept):
    r"""The multilinear extension approximation of the Shapley value in a weighted
    voting game using the technique proposed by Owen. For details see this paper: 
    `"Multilinear Extensions of Games."  <https://www.jstor.org/stable/2661445#metadata_info_tab_contents>`_
    """

    def __init__(self, epsilon: float=10**-8):
        self.epsilon = epsilon

    def _setup(self, W: np.ndarray):
        """Creating an empty Shapley value matrix."""
        self._Phi = np.zeros(W.shape)

    def _approximate(self, W: np.ndarray, q: float):
        """Using the naive multilinear approximation method."""
        for data_point in range(W.shape[0]):
            mu = float(np.mean(W[data_point,:]))
            var = float(np.var(W[data_point,:]))
            sigma_s = create_standard_deviation_vector(var, W.shape[1])
            for player_index in range(W.shape[1]):
                w = W[data_point, player_index]
                a_s, b_s = create_integrand_vectors(W.shape[1], q, w)
                shap_in_game = norm.cdf(b_s, loc=mu, scale=sigma_s)-norm.cdf(a_s, loc=mu, scale=sigma_s)
                self._Phi[data_point, player_index] += np.sum(shap_in_game)

        self._Phi = self._Phi / np.sum(self._Phi, axis=1).reshape(-1, 1)

    def solve_game(self, W: np.ndarray, q: float):
        r"""Solving the weigted voting game(s).

        Args:
            W (Numpy array): An :math:`n \times m` matrix of voting weights for the :math:`n` games with :math:`m` players.
            q (float): Quota in the games.
        """
        self._check_quota(q)
        self._setup(W)
        self._approximate(W, q)
        self._run_sanity_check(W, self._Phi)
        self._set_average_shapley()
        self._set_shapley_entropy()

    def get_solution(self) -> np.ndarray:
        r"""Returning the solution.

        Return Types:
            Phi (Numpy array): Approximate Shapley matrix of players in the game(s) with size :math:`n \times m`.
        """
        return self._Phi



from .exact_enumeration import ExactEnumeration
from .permutation_sampler import PermutationSampler
from .multilinear_extension import MultilinearExtension
from .expected_marginal_contributions import ExpectedMarginalContributions



"""Solution Concept base class."""

import numpy as np

class SolutionConcept(object):
    """Solution Concept base class with constructor and public methods."""

    def __init__(self):
        """Creating a Solution Concept."""
        pass

    def _check_quota(self, q: float):
        """Checking for negative quota."""
        assert 0.0 <= q

    def _verify_result_shape(self, W: np.ndarray, Phi: np.ndarray):
        """Checking the shape of the Shapley value matrix."""
        assert W.shape == Phi.shape

    def _verify_distribution(self, Phi: np.ndarray):
        """Verify distribution hypothesis."""
        assert np.sum(Phi) - Phi.shape[0] < 0.001
        

    def _run_sanity_check(self, W: np.ndarray, Phi: np.ndarray):
        """Checking the basic assumptions about the Shapley values."""
        self._verify_result_shape(W, Phi)
        self._verify_distribution(Phi)

    def _set_average_shapley(self):
        """Calculating the average Shapley value scores."""
        self._Phi_tilde = np.mean(self._Phi, axis=0)

    def _set_shapley_entropy(self):
        """Calculating the Shapley entropy score."""
        self._shapley_entropy = -np.sum(self._Phi_tilde*np.log(self._Phi_tilde))

    def get_average_shapley(self) -> np.ndarray:
        """Getting the average Shapley value scores."""
        return self._Phi_tilde

    def get_shapley_entropy(self) -> float:
        """Getting the Shapley entropy score."""
        return self._shapley_entropy



import os
from setuptools import find_packages, setup

on_rtd = os.environ.get("READTHEDOCS") == "True"

install_requires = ["numpy",
                    "scipy",
                    "six"]

if not on_rtd:
    install_requires.append("numba")

setup_requires = ['pytest-runner']


tests_require = ["pytest",
                 "pytest-cov",
                 "mock"]


keywords = ["shapley",
            "random-forest",
            "data-science",
            "shap",
            "ensemble",
            "expert-system",
            "explanability",
            "voting-classifier",
            "classifier",
            "machine-learning",
            "deep-learning",
            "deeplearning"]


setup(
  name = "shapley",
  packages = find_packages(),
  version = "1.0.0",
  license = "MIT",
  description = "A general purpose library for quantify the value of classifiers in a machine learning ensemble.",
  author = "Benedek Rozemberczki",
  author_email = "benedek.rozemberczki@gmail.com",
  url = "https://github.com/benedekrozemberczki/shapley",
  download_url = "https://github.com/benedekrozemberczki/shapley/archive/v_10000.tar.gz",
  keywords = keywords,
  install_requires = install_requires,
  setup_requires = setup_requires,
  tests_require = tests_require,
  classifiers = ["Development Status :: 3 - Alpha",
                 "Intended Audience :: Developers",
                 "Topic :: Software Development :: Build Tools",
                 "License :: OSI Approved :: MIT License",
                 "Programming Language :: Python :: 3.7"],
)



import random
import numpy as np
from scipy.stats import norm
from shapley.solution_concept import SolutionConcept

class MultilinearExtension(SolutionConcept):
    r"""The multilinear extension approximation of the Shapley value in a weighted
    voting game using the technique proposed by Owen. For details see this paper: 
    `"Multilinear Extensions of Games."  <https://www.jstor.org/stable/2661445#metadata_info_tab_contents>`_
    """
    def _setup(self, W: np.ndarray):
        """Creating an empty Shapley value matrix."""
        self._Phi = np.zeros(W.shape)

    def _approximate(self, W: np.ndarray, q: float):
        """Using the naive multilinear approximation method."""
        mu = np.tile(np.mean(W, axis=1).reshape(-1, 1), W.shape[1])
        var = np.tile(np.var(W, axis=1).reshape(-1, 1), W.shape[1])
        upper = (np.zeros(W.shape) + q - mu)/var
        lower = (np.zeros(W.shape) + q - W - mu)/var
        self._Phi = norm.cdf(upper, 0, 1) - norm.cdf(lower, 0, 1)
        self._Phi = self._Phi / np.sum(self._Phi, axis=1).reshape(-1, 1)

    def solve_game(self, W: np.ndarray, q: float):
        r"""Solving the weigted voting game(s).

        Args:
            W (Numpy array): An :math:`n \times m` matrix of voting weights for the :math:`n` games with :math:`m` players.
            q (float): Quota in the games.
        """
        self._check_quota(q)
        self._setup(W)
        self._approximate(W, q)
        self._run_sanity_check(W, self._Phi)
        self._set_average_shapley()
        self._set_shapley_entropy()

    def get_solution(self) -> np.ndarray:
        r"""Returning the solution.

        Return Types:
            Phi (Numpy array): Approximate Shapley matrix of players in the game(s) with size :math:`n \times m`.
        """
        return self._Phi
